/**************************************************************************
 * SAMSUNG 11Mbps Wireless LAN PCMCIA device driver for Linux (swld11_cs).
 *
 * Copyright (c) 2000 SAMSUNG Electro-Mechanics CO.,Ltd. -- http://www.MagicLan.com
 * All rights reserved.
 *
 *
 *	See README.swld11_cs for more information.
 */

/***********************************************************************************************
  Header : semmain_cs.h 28 Feb 2000 by J.J.LEE
 * 
 *    Rev 1.0   28 Feb 2000 20:32:32   J.J.LEE
 *      Initial revision.

************************************************************************************************/
//*******************************************************************************************
#include <asm/io.h>

//#define CNV_LITTLE_TO_INT(x)			// No endianess conversion needed										

#define FAR							// flat 32-bits code
#define BASED

typedef unsigned char				BYTE;
typedef unsigned short			     WORD;
typedef unsigned long				DWORD;


#define IN_PORT_BYTE(port)		((BYTE)inb( (WORD)(port) ))
#define IN_PORT_WORD(port)		((WORD)inw( (WORD)(port) ))
#define OUT_PORT_BYTE(port, value)	(outb( (BYTE) (value), (WORD)(port) ))
#define OUT_PORT_WORD(port, value)	(outw( (WORD) (value), (WORD)(port) ))

#define IN_PORT_STRING( port, src, n)	(insw( (WORD)(port),src,n))
#define OUT_PORT_STRING( port, dst, n)	(outsw((WORD)(port),dst,n))

//#endif	/* LINUX */

#ifdef __cplusplus
#define EXTERN_C extern "C"
#else
#define EXTERN_C
#endif //__cplusplus


#define TRUE  1
#define FALSE 0
#define ESSID_MAX_SIZE	32
#define MAX_DATALEN 2304    
#define SEM_LENGTH_OF_ADDRESS 6

#define INI_TICK_INI 0x20000L


//the Status of Card.
#define CARD_STAT_PRESENT 			0x8000

#define SWAPBYTE(S) (((S & 0X00FF)<<8) | ((S & 0xFF00) >> 8))

#define SemDisableInterrupt(Adapter) \
              semPutIOPortWord(Adapter,INTEN,0x00);

#define SemEnableInterrupt(Adapter) \
              semPutIOPortWord(Adapter, INTEN, INFDROP | INFOEN  | RXEN | ALLOCEN);

//for implementing WEP function.
#define WEP_64BIT             3
#define WEP_128BIT            7
#define WEP_64BIT_KEYLEN      5
#define WEP_128BIT_KEYLEN     13
#define WEP_64BIT_FULLKEYLEN  8
#define WEP_128BIT_FULLKEYLEN 16

//#define MAX_KEY_LEN (2+13*2)
#define CRC32_POLY_LE 0xedb88320UL
#define WEP_FULLKEYLEN 8
#define WEP_SBOXLEN 256
#define WEP_SWAP(a,b) { BYTE c; c = (a); (a) = (b);(b) = (c);}

typedef struct wepprng
{
  BYTE s[WEP_SBOXLEN];
  WORD i;
  WORD j;
}wepprng_t;

#define WEP_IV_LEN  4
#define WEP_ICV_LEN 4

#define FCC   0x10 //channel 1 ~ 11
#define ETSI  0x30 //channel 1 ~ 13
#define SPAIN 0x31 //channel 10~ 11
#define FRANCE 0x32 //channel 10~13

/********************************************************************************
*		Transmit Frame Structure definition
********************************************************************************/


typedef struct  
{	
	WORD Status;
	WORD Reserved1[2];
	WORD SwSupport[2];
	WORD Reserved2;
	WORD TxControl;
} tx_Control, *ptx_Control;

typedef struct
{
	WORD FrameControl;
	WORD DurationID;
	BYTE Address1[6];
	BYTE Address2[6];
	BYTE Address3[6];
	WORD SequenceControl;
	BYTE Address4[6];
	WORD DataLen;
} _802_11_header, *p802_11_header;

typedef struct
{
	BYTE DestAddr[6];
	BYTE SrcAddr[6];
	WORD Length;
} _802_3_header, *p802_3_header;
	

typedef struct 
{
	tx_Control t_tx_control;
	_802_11_header t_802_11_header;
	_802_3_header t_802_3_header;
	BYTE data[MAX_DATALEN];
} tx_frame, *pTX_Frame;	
	

typedef struct _FID_STRUCT
{
 	struct _FID_STRUCT* next;
	WORD fid;
}FID_STRUCT,*PFID_STRUCT;
	
// TxControl field    119P

#define TXOK 		0X0002
#define TXEX 		0X0004
#define STRUCTYPE	0X0018
#define T_MACPORT	0X0700  

#define FORMAT802_3  0x00
#define FORMAT802_11 0x08
#define TODS         0x0100
#define FROMDS       0x0200
// Status field

#define RETRYERR	0X0001
#define AGEDERR		0X0002
#define DISCON		0X0004
#define FROMERR		0X0008

	
//*************************************************************************
//   Record Structure Format
//**************************************************************************

typedef struct {
	WORD RecordLen;
        WORD RID;
	BYTE Data[250];
} Record_Buffer, *pRecord_Buffer;
	
/********************************************************************************
*		Receive Frame Structure definition
********************************************************************************/

typedef struct
{
	WORD Status;
	WORD Reserved1[2];
	BYTE Signal;
	BYTE Silence;
	BYTE RxFlow;
	BYTE Rate;
	WORD Reserved2[2];
} rx_Control, *prx_Control;

typedef struct
{
	rx_Control r_rx_control; 
	_802_11_header r_802_11_header;
	_802_3_header r_802_3_header;
	BYTE data[MAX_DATALEN];
} rx_frame, *pRX_Frame;



//Status field    125P

#define FCSERR		0X0001
#define UNDECR		0X0002
#define S_MACPORT	0X0700  
#define MESSAGETYPE	0Xe000


//RX message Type.
#define Rx_STATUS_TYPE_NORMAL         0x0000
#define RX_STATUS_TYPE_RFC1042        0x2000
#define RX_STATUS_TYPE_BRIDGE_TUNNEL  0x4000
#define RX_STATUS_TYPE_WMP            0x6000

typedef enum {
	RX_INT      = 0x01,
	TX_INT      = 0x02,
	TXEXC_INT   = 0x03,
	ALLOC_INT   = 0x04,
//	CMD_INT     = 0x05,
	INFO_INT    = 0x08,
	INFDROP_INT = 0x0E,
//	WTERR_INT   = 0x0F,
	TICK_INT    = 0x10,
	UNKNOWN     = 0x20
	} INTERRUPT_TYPE;
 
#define GetInterruptType(_A,_I)    \
        (_I & RX)     ? RX_INT :    \
        (_I & TX)     ? TX_INT :    \
        (_I & TXEXC)  ? TXEXC_INT : \
        (_I & ALLOC)  ? ALLOC_INT : \
        (_I & INFO)   ? INFO_INT  : \
        (_I & INFDROP)? INFDROP_INT :\
        (_I & TICK)   ? TICK_INT : UNKNOWN
/********************************************************************************
*		     Register Table definition
********************************************************************************/

// Command / Status
#define COMMAND 	0x00
#define PARAM0	 	0x02
#define PARAM1  	0x04
#define PARAM2  	0x06
#define STATUS  	0x08
#define RESP0  		0x0a
#define RESP1		0x0c
#define RESP2		0x0e

//  FID Management
#define INFOFID		0X10
#define RXFID		0x20
#define ALLOCFID	0x22
#define TXCOMPLFID	0x24

// BAP 0
#define SELECT0		0x18
#define OFFSET0		0x1c
#define DATA0		0x36

// BAP 1
#define SELECT1		0x1a
#define OFFSET1		0x1e
#define DATA1		0x38

// Event
#define EVSTAT		0x30
#define INTEN		0x32
#define EVACK		0x34

// Control
#define CONTROL		0x14

// Host Software
#define SWSUPPORT0	0x28
#define SWSUPPORT1	0x2a
#define SWSUPPORT2	0x2c

// AUX Port
#define AUX_BASE         0x3A
#define AUX_OFFSET       0x3C
#define AUX_DATA         0x3E

/********************************************************************************
*		Description of Register Bits 
********************************************************************************/

//Command /status    30P

#define CMDCODE		0x003f
#define C_MACPORT	0x0700
#define C_BUSY		0x1000   
#define RECL		0x0100
#define WRITE_ACCESS	0x0100
#define PROGMODE	0x0300

#define COMMANDPARM	0xffff

#define CMDCODE		0x003f
#define RESULT		0x7f00

#define COMMANDRESP	0xffff

// COMAND Result Code
#define COMMAND_SUCCESS  0x00
#define CARD_FAILURE     0x01
#define NO_BUFFER_SPACE  0x05
#define COMMAND_ERROR    0x7F

// FID Management

#define INFORMATONFID	0xffff
#define RECEIVEDFID	0xffff
#define ALLOCATEDFID	0xffff
#define COMPLETEDTXFID	0xffff

 
// BAP   16P

#define FID_RID		0xffff
#define DATAOFFSET	0x0ffe
#define ERR		0x4000
#define B_BUSY		0x8000
#define DATA		0xffff

// Evstat

#define RX		0x0001
#define TX		0x0002
#define TXEXC		0x0004
#define ALLOC		0x0008
#define CMD		0x0010
#define INFO		0x0080
#define INFDROP		0x2000
#define WTERR		0x4000
#define TICK		0x8000

// IntEn

#define RXEN		0x0001
#define TXEN		0x0002
#define TXEXCEN		0x0004
#define ALLOCEN		0x0008
#define CMDEN		0x0010
#define INFOEN		0x0080
#define INFDROPEN	0x2000
#define WTERREN		0x4000
#define TICKEN		0x8000

// EvAck

#define RXACK		0x0001
#define TXACK		0x0002
#define TXEXCACK	0x0004
#define ALLOCACK	0x0008
#define CMDACK		0x0010
#define INFOACK		0x0080
#define INFDROPACK	0x2000
#define WTERRACK	0x4000
#define TICKACK		0x8000

// Control

#define AUX_ENABLE  0x8000 // Enable auxiliary port access.
#define AUX_DISABLE 0x4000
#define AUX_OPEN   0xC000
#define AUX_CLOSE  0x0000

#define AUX_PARAM0_KEY   0xFE01
#define AUX_PARAM1_KEY   0xDC23
#define AUX_PARAM2_KEY   0xBA45

// Host Software

#define SOFT_SUPPORT_VAL 	0xffff


/********************************************************************************
*		     Command Table definition
********************************************************************************/

// Controller/ State

#define INITIALIZE	0x00
#define ENABLE		0x01
#define DISABLE		0x02
#define DIAGNOSE	0x03

// Buffer Management

#define ALLOCATE	0x0a
#define TRANSMIT	0x0b

// Regulate

#define NOTIFY		0x10
#define INQUIRE		0x11

// Configure

#define ACCESS		0x21
#define DOWNLOAD	0x22


#define ACCESS_WRITE  0x0100
#define ACCESS_READ   0x0000
/********************************************************************************
*		     Configuration Records definition
********************************************************************************/


// Network Parameters, static Configuration Entities

#define CNFPORTTYPE		0XFC00
#define CNFOWNMACADDRESS	0XFC01
#define CNFDESIREDSSID		0XFC02
#define CNFOWNCHANNEL		0XFC03
#define CNFOWNSSID		0XFC04
#define CNFOWNATIMWINDOW	0XFC05
#define CNFSYSTEMSCALE		0XFC06
#define CNFMAXDATALENGTH	0XFC07
#define CNFWDSADDRESS		0XFC08
#define CNFPMENABLED		0XFC09
#define CNFPMEPS		0XFC0A
#define CNFMULTICASTRECEIVE	0XFC0B
#define CNFMAXSLEEPDURATION	0XFC0C
#define CNFPMHOLDOVERDURATION	0XFC0D
#define CNFOWNNAME		0XFC0E
#define CNFOWNDTIMPERIOD	0XFC10
#define CNFWDSADDRESS1		0XFC11		
#define CNFWDSADDRESS2		0XFC12
#define CNFWDSADDRESS3		0XFC13
#define CNFWDSADDRESS4		0XFC14
#define CNFWDSADDRESS5		0XFC15
#define CNFWDSADDRESS6		0XFC16
#define CNFMULTICASTPMBUFFERING 0XFC17

//SEM 10/20/99 ver 2.1 by jjlee for supporting WEP.
#define CNFWEPDEFAULTKEYID  0xFC23
#define CNFDEAULTKEY0       0xFC24
#define CNFDEAULTKEY1       0xFC25  
#define CNFDEAULTKEY2       0xFC26
#define CNFDEAULTKEY3       0xFC27
#define CNFWEPFLAGS         0xFC28
#define CNFWEPKEYMAPPINGTABLE 0xFC29
#define CNFAUTHENTICATION   0xFC2A  
//end of sem

// NetworkParameters, Dynamic Configuration Entities

#define GROUPADDRESS		0XFC80
#define CREATEIBSS		0XFC81
#define FRAGMENTATIONTHRESHOLD	0XFC82
#define RTSTHRESHOLD		0XFC83
#define TXRATECONTROL		0XFC84
#define PROMISCUOUSMODE		0XFC85
#define FRAGMENTATIONTHRESHOLD0	0XFC90
#define FRAGMENTATIONTHRESHOLD1	0XFC91
#define FRAGMENTATIONTHRESHOLD2	0XFC92
#define FRAGMENTATIONTHRESHOLD3	0XFC93
#define FRAGMENTATIONTHRESHOLD4	0XFC94
#define FRAGMENTATIONTHRESHOLD5	0XFC95
#define FRAGMENTATIONTHRESHOLD6	0XFC96
#define RTSTHRESHOLD0		0XFC97 	
#define RTSTHRESHOLD1		0XFC98 
#define RTSTHRESHOLD2		0XFC99 
#define RTSTHRESHOLD3		0XFC9A 
#define RTSTHRESHOLD4		0XFC9B 
#define RTSTHRESHOLD5		0XFC9C 
#define RTSTHRESHOLD6		0XFC9D
#define TXRATECONTROL0		0XFC9E
#define TXRATECONTROL1		0XFC9F
#define TXRATECONTROL2		0XFCA0
#define TXRATECONTROL3		0XFCA1
#define TXRATECONTROL4		0XFCA2
#define TXRATECONTROL5		0XFCA3
#define TXRATECONTROL6		0XFCA4

//SEM 2000/02/02 VER 3.3 for including the API.
#define cnfShortPreamble       0XFCB0


// Beheavior Parameters

#define TICKTIME		0XFCE0



/********************************************************************************
*		     Information Records definition
********************************************************************************/

// NIC Information    71P

#define MAXLOADTIME		0XFD00
#define DOWNLOADBUFFER		0XFD01
#define PRIIDENTITY		0XFD02
#define PRISUPRANGE		0XFD03
#define CFIACTRANGES		0XFD04
#define NICSERIALNUMBER		0XFD0A
#define NICIDENTITY		0XFD0B
#define MFISUPRANGE		0XFD0C
#define CFISUPRANGE		0XFD0D
#define CHANNELLIST		0XFD10
#define REGULATORYDOMAINS	0XFD11
#define TEMPTYPE		0XFD12
#define CIS			0XFD13
#define STAIDENTITY		0XFD20
#define STASUPRANGE		0XFD21
#define MFIACTRANGES		0XFD22
//#define CFIACTRANGES		0xFD23   

// MAC Information

#define PORTSTATUS		0xFD40
#define CURRENTSSID		0XFD41
#define CURRENTBSSID		0xfd42
#define COMMSQUALITY		0XFD43
#define CURRENTTXRATE		0XFD44
#define CURRENTBEACONINTERVAL	0XFD45
#define CURRENTSCALETHRESHOLDS	0XFD46
#define PROTOCOLRSPTIME		0XFD47
#define SHORTRETRYLIMIT		0XFD48
#define LONGRETRYLIMIT		0XFD49
#define MAXTRANSMITLIFETIME	0XFD4A
#define MAXRECEIVELIFETIME	0XFD4B
#define CFPOLLABLE		0XFD4C
#define AUTHENTICATIONALGORITHMS		0XFD4D
#define PRIVACYOPTIONIMPLEMENTED		0XFD4F
#define CURRENTTXRATE1		0XFD80
#define CURRENTTXRATE2		0XFD81
#define CURRENTTXRATE3		0XFD82
#define CURRENTTXRATE4		0XFD83
#define CURRENTTXRATE5		0XFD84
#define CURRENTTXRATE6		0XFD85
#define OWNMACADDRESS		0XFD86
#define PCFINFO			0XFD87

// Modem Information

#define PHYTYPE			0XFDC0
#define CURRENTCHANNEL 		0XFDC1
#define CURRENTPOWERSTATE	0XFDC2
#define CCAMODE			0XFDC3
#define SUPPORTEDDATARATES	0XFDC6
/********************************************************************************
*		     Information Frame Structure definition
********************************************************************************/

// MAC Management

#define HANDOVERADDRESS		0XF000
#define LINKSTATUS		0XF200
#define ASSOCIATIONSTATUS	0XF201

// Diagnose

#define COMMUNICATIONTALLIES	0XF100
#define SCANRESULTS		0XF101


// CommunicationTallies

#define FRAMELEN		0X00
#define INFOTYPE		0X01
#define TXUNICASTFRAMES		0X02
#define TXMULTICASTFRAMES	0X03
#define TXFRAGMENTS		0X04
#define TXUNICASTOCTETS		0X05
#define TXMULTICASTOCTETS	0X06
#define TXDEFERREDTRANSMISSIONS 0X07
#define TXSINGLERETRYFRAMES	0X08
#define TXMULTIPLERETRYFRAMES   0X09
#define TXRETRYLIMITEXCEEDED	0X0A
#define TXDISCARDS		0X0B
#define RXUNICASTFRAMES		0X0C
#define RXMULTICASTFRAMES	0X0D
#define RXFRAGMENTS		0X0E
#define RXUNICASTOCTETS		0X10
#define RXMULTICASTOCTETS	0X11
#define RxFCSErrors		0x12
#define RXDISCARDSNOBUFFER	0X13
#define TXDISCARDSWRONGSA	0X14
#define RXDISCARDSWEPUNDECRYPTABLE	0X15
#define RXMESSAGEINMSGFRAGMENTS		0X16
#define RXMESSAGEINBADMSGFRAGMENTS	0X17

// ScanResult subfields

#define SRTCHID		0X00
#define SRTANL		0X01
#define SRTSL		0X02
#define	SRTBSSID	0X03
#define SRTBcnInt	0x06
#define SRTCAP		0X07
#define SRTSSID		0x08

// SRTCAP

#define ESS		      0X0001
#define IBSS		  0X0002
#define CFPOLLble	  0x0004
#define CFPOLLREQUEST 0X0008
#define PRIVACY		  0X0010

//Link Status 
#define NOT_CONNECTED  0x00
#define CONNECTED      0x01
#define DISCONNECTED   0x02
#define AP_CHANGE      0x03
#define AP_OUTOF_RANGE 0x04
#define AP_IN_RANGE    0x05
#define ASSOCIATION_FAILED 0x06

//Assoication Status
#define STA_ASSOCIATED    0x01
#define STA_REASSOCIATED  0x02
#define STA_DISASSOCIATED 0x03


// Adapter data for netdevice
typedef struct _SEM_ADAPTER {
	
//resource information
  	WORD   IoBaseAddrR;
	BYTE	  InterruptNumber;

     BYTE   Desired_ESSID[32+1];        // ESSID value
 	BYTE   Desired_ESSID_Length;     // ESSID length
 	BYTE   StationName[32+1];          // Station Name
 	BYTE   StationNameLength;        // Station Name Length     
 	BYTE   NetworkAddress[6];        // User defined MAC ID.
 
 //Current MAC ID usded by driver.mostly use the MACID in the firmware,
 // if defined in the registry,then uses the MACID in the registr.
 	BYTE   CurrentNetworkAddress[6]; 
 
 // RTS Threshold (3000:0ff, 500:Hidden station)
 	WORD  Threshold;
 
  //1:Fixed 1Mb/s,2:Fixed 2Mb/s,3:Auto Rate(1,2Mb/s),4:Fixed 5.5Mbit/s,
  //8:Fixed 11Mbit/s,15: Auto fully
 	WORD  TransmitRate;        			  			
 	WORD  Channel;                  // Channel (it used at Ad_Hoc Mode)
 //for supporting Short Preamble and Fragmentation.
 	WORD ShortPreamble;
 	WORD FragThresh; 	
 	WORD  PSModeValue;              //0:Disabled, 1:Enabled.
// 	DWORD  PSLevel;		        // Power management level(0:Normal,1:Enhanced).	 				
 	WORD  NetworkMode;              //3:Ad_Hoc , 1:Infrastructure.
 	WORD  APDensity;	            //1:Low Density, 2:Medium Density, 3:High Density
 	
 	WORD  CurrentChannel;           // Currently used Channel.
 //the following value is connected with the CurrentReg3 value.
 //so it checks the correct use of channel by each country.  
 //Allowed to be used with this NIC.
 	WORD   AvailableChannel;         
 	WORD  CurrentRegDomain;         // Current Regulatory Domain.
 	
 //Wep Required(0:Disabled,3:64bit WEP,7:128bit WEP)
 	WORD  WepRequired;
 	WORD  AvailableWEP;		  //available WEP..
 	WORD  WepDefaultKeyId;          // Default WEP KEY ID.	
 // Default Keys.
 	BYTE   WepDefaultKey0[14+1];
 	BYTE   WepDefaultKey1[14+1];
 	BYTE   WepDefaultKey2[14+1];
 	BYTE   WepDefaultKey3[14+1];

     BYTE   iv[4];
     WORD   ivCount;
     WORD   ivCountMax;
     BYTE   key[16];
     BYTE   rx_key[16];
     BYTE   tx_wepBox[2348];
     BYTE   rx_wepBox[2348];

 	wepprng_t prng;                // RC4 data structure.
 	DWORD	crc32_table[256];

 	BYTE  EncryptInDriver;
 	BYTE  DecryptInDriver;
 	BYTE  reassemblyInProgress;
     WORD  savePos ;
 	
 	WORD  AuthenticationType;      // 0: open system, 1:Shared key.
 	BYTE   PromiscuousMode;         //  Control ifPromiscuous mode reception is enabled
	BYTE   CurrentBSSID[6];          //In ESS,connected AP MAC ID.
  
// 	BYTE   SendTempBuffer[3000];     // TX Buffer.
	tx_frame TxFrame;
	WORD	FIDValue;
	BYTE    reclaimStatus;
	rx_frame RxFrame;

 	WORD   PacketLength;
 	WORD   NumberOfAddress;
 //	BYTE   MulticastAddress[SEM_MAX_MULTICASTLIST][SEM_LENGTH_OF_ADDRESS]; 
 	rx_Control RXControl;            // recieve RX frame Control field.
 	WORD   InterruptStatus;
// For statistic use.
 	WORD   LinkStatus;
//	 NDIS_MEDIA_STATE   LinkIsActive;
 
 	WORD   AssociationStatus;
 	WORD   PortStatus; 	          // 1 ~ 6.
 	BYTE   CurrentSSID[32];          // Current BSSID
 	BYTE   CurrentSSIDLength;
 	WORD   CQcurrBSS; 	          // Communication Quality.
 	WORD   ASLcurrBSS; 	          // Average Signal Level.
 	WORD   ANLcurrFC; 	          // Average Noise Level.
 	WORD   CurrentTxRate;            // current Tx Rate.
 	WORD   CurrentPowerState;        // current power state.

//for Firmware Version display
 	WORD PRIVariant;
 	WORD PRIMajorVersion;
 	WORD PRIMinorVersion;
 
 	WORD STAVariant;
 	WORD STAMajorVersion;
 	WORD STAMinorVersion;  
 
     WORD CardStat;
     DWORD TickIni;

     BYTE scan_buffer[1000];
     BYTE   old_Desired_ESSID[32+1];        // ESSID value
 	BYTE   old_Desired_ESSID_Length;     // ESSID length
     WORD   old_NetworkMode;              //3:Ad_Hoc , 1:Infrastructure.	

} SEM_ADAPTER, *PSEM_ADAPTER;


// Shortcuts
#ifdef WIRELESS_EXT
typedef struct iw_statistics	iw_stats;
typedef struct iw_quality	iw_qual;
typedef struct iw_freq		iw_freq;
#endif /* WIRELESS_EXT */


WORD semGetIOPortWord(PSEM_ADAPTER Adapter,WORD port);
void semPutIOPortWord(PSEM_ADAPTER Adapter,WORD port,WORD value);
WORD issueCmd0(PSEM_ADAPTER Adapter,WORD CmdType);
WORD issueCmd1(PSEM_ADAPTER Adapter,WORD CmdType,WORD p0);
WORD issueCmd2(PSEM_ADAPTER Adapter,WORD CmdType,WORD p0,WORD p1);
WORD issueCmd3(PSEM_ADAPTER Adapter,WORD CmdType,WORD p0,WORD p1,WORD p2);
void issueEventACK(PSEM_ADAPTER Adapter,WORD eventACKType);

WORD cmdTime(WORD cmd);
WORD waitbusy(PSEM_ADAPTER Adapter,WORD cmdTypePort);
WORD waitevent(PSEM_ADAPTER Adapter,WORD eventType);
WORD getCommandResult(PSEM_ADAPTER Adapter);
void issueEventACK(PSEM_ADAPTER Adapter,WORD eventACKType);

WORD Initialize(PSEM_ADAPTER Adapter);
WORD Enable(PSEM_ADAPTER Adapter);
WORD Disable(PSEM_ADAPTER Adapter);

WORD cardStatusRead(PSEM_ADAPTER Adapter);
WORD cardInitialSet(PSEM_ADAPTER Adapter);

WORD getNetworkMode(PSEM_ADAPTER Adapter,WORD* value);
WORD setNetworkMode(PSEM_ADAPTER Adapter,WORD value);

WORD getESSID(PSEM_ADAPTER Adapter,BYTE* byteString,WORD* Length);
WORD setESSID(PSEM_ADAPTER Adapter,BYTE* byteString,WORD Length);
WORD getSSID(PSEM_ADAPTER Adapter,BYTE* byteString,WORD* Length);
WORD setSSID(PSEM_ADAPTER Adapter,BYTE* byteString,WORD Length);

WORD getChannel(PSEM_ADAPTER Adapter,WORD* value);
WORD setChannel(PSEM_ADAPTER Adapter,WORD value);

WORD getAPDensity(PSEM_ADAPTER Adapter,WORD* value);
WORD setAPDensity(PSEM_ADAPTER Adapter,WORD value);

WORD getPwrMgmt(PSEM_ADAPTER Adapter,WORD* PSEnable);
WORD getPwrMgmtEnhance(PSEM_ADAPTER Adapter,WORD* PSEhance);
WORD setPwrMgmt(PSEM_ADAPTER Adapter,WORD PSEnable);//,WORD PSEnhance);

WORD getStationName(PSEM_ADAPTER Adapter,BYTE* byteString,WORD* Length);
WORD setStationName(PSEM_ADAPTER Adapter,BYTE* byteString,WORD Length);

WORD getRTSThreshold(PSEM_ADAPTER Adapter,WORD* value);
WORD setRTSThreshold(PSEM_ADAPTER Adapter,WORD value);

WORD getTxRate(PSEM_ADAPTER Adpater,WORD* value);
WORD setTxRate(PSEM_ADAPTER Adapter,WORD value);

WORD getCurrentESSID(PSEM_ADAPTER Adapter,BYTE* byteString,WORD* Length);
WORD getPortStatus(PSEM_ADAPTER Adapter,WORD* value);
WORD getCurrentChannel(PSEM_ADAPTER Adapter,WORD* value);
WORD getCurrentBSSID(PSEM_ADAPTER Adapter,BYTE* byteString);
WORD getCommsQuality(PSEM_ADAPTER Adapter,WORD* wordString );
WORD getCurrentTxRate(PSEM_ADAPTER Adapter,WORD* value);
WORD getCurrentPwrState(PSEM_ADAPTER Adapter,WORD* value);

WORD getRecordFrame(PSEM_ADAPTER Adapter,WORD* pBuffer);
WORD putRecordFrame(PSEM_ADAPTER Adapter,WORD* pBuffer);

WORD getFrame0(PSEM_ADAPTER Adapter,WORD RIDValue,WORD* dst,WORD Offset,WORD Length);
WORD getFrame1(PSEM_ADAPTER Adapter,WORD RIDValue,WORD* dst,WORD Offset,WORD Length);
WORD putFrame0(PSEM_ADAPTER Adapter,WORD RIDValue, WORD* src,WORD Offset,WORD Length);
WORD putFrame1(PSEM_ADAPTER Adapter,WORD RIDValue, WORD* src,WORD Offset,WORD Length);
WORD cardInitialSet(PSEM_ADAPTER Adapter);
WORD cardStatusRead(PSEM_ADAPTER Adapter);
WORD cardReset(PSEM_ADAPTER Adapter);
WORD getInfoFrame(PSEM_ADAPTER Adapter,WORD InfoFID);
WORD getShortPreamble(PSEM_ADAPTER Adapter,WORD* value);
WORD setShortPreamble(PSEM_ADAPTER Adapter,WORD value);
WORD getFragmentationThreshold(PSEM_ADAPTER Adapter,WORD* value);
WORD setFragmentationThreshold(PSEM_ADAPTER Adapter,WORD value);
WORD calibrate(PSEM_ADAPTER Adapter);
WORD sendTxFrame(PSEM_ADAPTER Adapter,BYTE* pBuffer,WORD Length,WORD* DataSended);
WORD semReceive(PSEM_ADAPTER Adapter,BYTE* pBuffer,WORD* Length);

WORD setWEPDefaultKeyID(PSEM_ADAPTER Adapter,WORD value);
WORD getWEPDefaultKeyID(PSEM_ADAPTER Adapter,WORD* value);

WORD setWEPDefaultKey0(PSEM_ADAPTER Adapter,BYTE* byteString);
WORD setWEPDefaultKey1(PSEM_ADAPTER Adapter,BYTE* byteString);
WORD setWEPDefaultKey2(PSEM_ADAPTER Adapter,BYTE* byteString);
WORD setWEPDefaultKey3(PSEM_ADAPTER Adapter,BYTE* byteString);

WORD getWEPFlags(PSEM_ADAPTER Adapter,WORD* value);
WORD setWEPFlags(PSEM_ADAPTER Adapter,WORD value);

WORD getAuthentication(PSEM_ADAPTER Adapter,WORD* value);
void initializewep(PSEM_ADAPTER Adapter);

WORD setAuthentication(PSEM_ADAPTER Adapter,WORD value);
WORD semstring2hex(char* ks,char* key);

DWORD crc32_block(PSEM_ADAPTER Adapter,DWORD reg, BYTE *p, WORD len);
DWORD crc32_blockfinish(DWORD reg);
void crc32_mktable(PSEM_ADAPTER Adapter);
void wep_initprng(PSEM_ADAPTER Adapter,BYTE *k,  WORD klen);
void wep_block_init(PSEM_ADAPTER Adapter, BYTE *d);
BYTE wep_nextprn(PSEM_ADAPTER Adapter);
void siteSurvey_start(PSEM_ADAPTER Adapter);
void siteSurvey_end(PSEM_ADAPTER Adapter);
void siteSurvey_cont(PSEM_ADAPTER Adapter);
void wep_encrypt(PSEM_ADAPTER Adapter,BYTE *data,WORD dlen);
void wep_decrypt(PSEM_ADAPTER Adapter,BYTE *data,WORD dlen);
